#!/usr/bin/perl

#
# Licensed Materials - Property of IBM
# Rational ClearCase
# (C) Copyright IBM Corp. 2014.  All Rights Reserved.
# US Government Users Restricted Rights -
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp
#


# This script is intended to assist in the migration from the 
# Base ClearCase/ClearQuest V2 Perl Integration to the 
# Change Management Integration (CMI).
#
# It reads the config.pl used by the Perl Integration, prompts
# for additional informaiton used by CMI, sets up attribute types
# used by CMI on the VOB, sets up CMI providers on the VOB and Branch
# Types, and creates a list of Versions to be linked to ClearQuest
# records by the CMI Repair Tool (clmutil).  Completing the migration requires
# running the CMI Repair Tool (clmutil) after this script has finished.
#
# This process is done per VOB.  If multiple VOBs are to be migrated
# then the script must be run for each VOB.

use strict;
use warnings;

our $Registry;
if ($^O =~ /Win32/) {
    require Win32::TieRegistry;
    import Win32::TieRegistry;
}
    

# This lets the print statements print before the die messages.
local $| = 1;

my $finish = 0;
my $vobinput = "";
my $vobresult = "";
my $pathseparator = "/";

if ( $^O eq 'MSWin32' ) { 
    $pathseparator = "\\";
}

print "\nWarning: It is recommended that the VOB be locked -nusers for the User running this script.";

while ( !$finish ) {
    # Prompt for VOB tag
    print "\nEnter the VOB tag to migrate or return to quit: ";
    # Get VOB tag from user
    $vobinput = <STDIN>;
    # Trim whitespace
    chomp $vobinput;
    # Exit if no VOB tag is provided
    exit 0 if ($vobinput eq "");
    # Run cleartool to verify VOB tag
    $vobresult = `cleartool lsvob $vobinput 2>&1`;

    # Check result
    # Not successful, continue loop
    if ( $vobresult =~ /cleartool: / && ${^CHILD_ERROR_NATIVE} != 0 ) {
	print "\nVOB tag not found.";
    } else {
	# Successful, we have a valid VOB tag
	print "\nVOB tag validated.";
	$finish = 1;
    }
}

# Get temp directory to use with the migration
my $tempdir;
print "\nEnter a temp directory for use with the migration or return to quit: ";
# Get temp dir from user
$tempdir = <STDIN>;
# Trim whitespace
chomp $tempdir;
# Exit if no dir is provided
exit 0 if ($tempdir eq "");

# Get file names or use defaults
my $tempinput = "";
my $file_name = "cmi_migration.txt";
my $versions_file_name = "cmi_migration_versions.txt";
print "\nEnter a file name to store migration configuration information or return to use the default($file_name): ";
$tempinput = <STDIN>;
chomp $tempinput;
if ( $tempinput ne "" ) {
    $file_name = $tempinput;
}
print "\nEnter a file name to store migration version information or return to use the default($versions_file_name): ";
$tempinput = <STDIN>;
chomp $tempinput;
if ( $tempinput ne "" ) {
    $versions_file_name = $tempinput;
}

# Setup file names to use with the migration
# Migration information will be stored in cmi_migration.txt
# Versions to migrate will be stored in cmi_migration_versions.txt
# to be used by the repair tool (clmutil) to make associations with the provider(s)
my $migration_file_name = "$tempdir" . "$pathseparator" . "$file_name";
my $migration_versions_file_name = "$tempdir" . "$pathseparator" . "$versions_file_name";


# Are they resuming a previous migration?
# If yes, skip to check if all details to setup CMI are present

# If the migration versions file already exists we need to see if it's non-empty
if ( -e $migration_versions_file_name ) {
    # Exists and is non-empty, quit
    if ( -s $migration_versions_file_name ) {
	die "\nMigration versions file $migration_versions_file_name exists and is non-empty." .
	"\n$migration_versions_file_name must point to an empty or non-existing file.";
    }
}

# If the migration configuration file already exists we will either resume or quit
if ( -e $migration_file_name ) { 
    # Exists and is non-empty
    if ( -s $migration_file_name ) {
	print "\nMigration configuration file $migration_file_name exists. Resume (y) or quit (n)?";
	my $resume = <STDIN>;
	chomp $resume;
	# If we're not resuming, quit
	if ( $resume ne "y" ) {
	    print "\nQuitting...\n";
	    exit 0; 
	}
	# If we're resuming, go to sub migrate and skip setup
	&migrate;
	exit 0;
    }
    # Exists and is empty is the same as not existing for us
}
# Does not exist yet
&setup;
&migrate;
exit 0;


sub setup {

print "\nChecking for the ClearCase directory...";

my $key = "";
my $configfile = "";
# Get the ClearCase directory
if ( $^O =~ /Win32/ ) { 
    # Windows
    # Check if this is 32 or 64 bit Windows
    my $OSbits = &winbitcheck();
    my $register;
    if ( $OSbits eq "64" ) {
	# Get the ClearCase directory from the 64-bit Registry entry
	$register = "HKEY_LOCAL_MACHINE\\SOFTWARE\\Wow6432Node\\Atria\\ClearCase\\2.0\\ProductHome";
    } elsif ( $OSbits eq "32" ) {
	# Get the ClearCase directory from the 32-bit Registry entry
	$register = "HKEY_LOCAL_MACHINE\\SOFTWARE\\Atria\\ClearCase\\2.0\\ProductHome";
    }
    # Grab the ClearCase directory
    $key = $Registry->{$register};
    $configfile = $key . "\\lib\\CQCC\\config.pl";

} else {
    # Unix
    $configfile = "/opt/rational/clearcase/lib/CQCC/config.pl";
}

# Check if config.pl exists
if ( -e $configfile ) {
    print "\nFound config.pl.";
} else {
    die "config.pl not found.";
}

print "\nExtracting information from config.pl...";

# Grab the CQ information we care about
open (input_file, "<", $configfile) or die $!;
my $lookingfordb = 0;
my @dblist;
my $tempindex = 0;
my $tempindex2 = 0;
my $islocal = 0;
my $dbset = "";
my $server = "";
my $serverroot = "";
my $ssl = "";
my $dbindex = 0;
my $subindex = 0;

while ( my $line = <input_file> ) {
    # Remove leading and trailing whitespace from $line
    $line =~ s/^\s+|\s+$//g;
    # CQ DB
    if ( $lookingfordb ) {
	# ) means the last DB entry
	$tempindex = index($line, "\)");
        if ( $tempindex != -1 ) {
	    # As long as it's not in a comment
            $tempindex2 = index($line, "#");
            if ( ($tempindex2 > $tempindex) || ($tempindex2 == -1) ) {
		# We're done looking for DBs
                $lookingfordb = 0;
            }
        }
        # We're looking for DBs so let's get the current one
        if ( $line =~ /:/ ) {
            # Find where the DB name ends and starts
            # Ignore the record type as we can't do anything with it
            $dbindex = index($line, ":");
            $subindex = rindex($line, "\"", $dbindex);
            if ( $subindex != -1 ) {
                push (@dblist, (substr $line, $subindex+1, ($dbindex - $subindex)-1));
            }
        }
    }
    elsif ( $line =~ /&SetConfigParm\("CQCC_DATABASE_ENTITY_LIST"/ ) {
	$lookingfordb = 1;
        # Check to see if there's a DB entry on the same line
        if ( $line =~ /:/ ) {
            # Get the number of DB entries if they're all on one line
            my $count = () = $line =~ /:/g;
            $dbindex = index($line, ":");
            $subindex = 0;
            # Get each DB entry
            # Note that each DB entry should be in the format "DB: record_type"
            for ( my $i = 0; $i < $count; $i++ ) {
                # Go back from the : to find the " that signifies the start of an entry
                $subindex = rindex($line, "\"", $dbindex);
                # This check shouldn't be necessary but it can't hurt
                if ( $subindex != -1 ) {
                    # Store the DB name in @dblist
                    push (@dblist, (substr $line, $subindex+1, ($dbindex - $subindex)-1));
                }
                # Get the next entry (if there is one)
                $dbindex = index($line, ":", $dbindex+1);
            }
    	}
    }
    # CQ DBSet
    elsif ( $line =~ /&SetConfigParm\("CQCC_DATABASE_SET"/ && (index($line, "#") != 0) ) {
	# Local CQ DBSet
	$tempindex = index($line, "\")");
	$tempindex2 = rindex($line, "\"", $tempindex-1);
	# If we already found a DBSet we're done.  Can't have more than one.
	if ( $dbset eq "" && ($tempindex2 > index($line, "#")) && ($tempindex != -1) && ($tempindex2 != -1) ) {
	    $dbset = substr $line, $tempindex2+1, ($tempindex - $tempindex2)-1;
	}
    }
    # CQ Web DBSet
    elsif ( $line =~ /&SetConfigParm\("CQCC_WEB_DATABASE_SET"/ && (index($line, "#") != 0) ) {
    	# Web CQ DBSet
	$tempindex = index($line, "\")");
 	$tempindex2 = rindex($line, "\"", $tempindex-1);
    	# If we already found a DBSet we're done.  Can't have more than one.
    	if ( $dbset eq "" && ($tempindex2 > index($line, "#")) && ($tempindex != -1) && ($tempindex2 != -1) ) {
	    $dbset = substr $line, $tempindex2+1, ($tempindex - $tempindex2)-1;
    	}
    }
    # CQ Web URL (server)
    elsif ( $line =~ /&SetConfigParm\("CQCC_SERVER"/ && (index($line, "#") != 0) ) {
	$tempindex = index($line, "\")");
	$tempindex2 = rindex($line, "\"", $tempindex-1);
	if ( ($tempindex2 > index($line, "#")) && ($tempindex != -1) && ($tempindex2 != -1) ) {
	    $server = substr $line, $tempindex2+1, ($tempindex - $tempindex2)-1;
	}
    }

    # SSL
    elsif ( $line =~ /&SetConfigParm\("CQCC_SERVER_SSL"/ && (index($line, "#") != 0) ) {
	$tempindex = index($line, "\")");
	$tempindex2 = rindex($line, "\"", $tempindex-1);
	if ( ($tempindex2 > index($line, "#")) && ($tempindex != -1) && ($tempindex2 != -1) ) {
	    $ssl = substr $line, $tempindex2+1, ($tempindex - $tempindex2)-1;
	}
    }
}
# At this point we have gathered the information we can from the VOB and config.pl

# Put together the server URL
my $serverurl = "";
if ( $server ne "" ) {
    $serverurl .= (($ssl eq "TRUE") ? "https://" : "http://");
    $serverurl .= $server;
    $serverurl .= "/cqweb/oslc";
} else {
    # Server was empty in config.pl
    # Prompt the user to provide one
    print "\nPlease enter a ClearQuest Web Server URL to use (e.g. http://cqserver.domain.com:12080): ";
    $serverurl = <STDIN>;
    chomp $serverurl;
    # If the last character is '/' we don't want to add another one
    if ( !(rindex($serverurl, "/") == (length $serverurl - 1)) ) {
	# Last character is NOT '/'
	$serverurl .= "/";
    }
    $serverurl .= "cqweb/oslc";
}

print "\nChecking trigger types for branch type restrictions...";

# Get the brtypes that have triggers on them
my %branchtypes;
$vobresult = `cleartool lstype -kind trtype -long -invob $vobinput 2>&1`;
if ( $vobresult =~ /cleartool: / && ${^CHILD_ERROR_NATIVE} != 0 ) {
    die "\nError listing trigger types on $vobinput.\n$vobresult";
}
else {
    my $checktriggers = 1;
    my $triggerindex = 0;
    my $branchindex = 0;
    my $onetrigger = 0;
    my $brtypecount = 0;
    my $tempbrpos = 0;
    my $brtype = "";
    my $endofbrtype = 0;
    while ( $checktriggers ) {
        # Check which trigger types are for the Perl integration
        # and see if they have branch type restrictions
        
        # Check for next Perl integration trigger type
        $triggerindex = index($vobresult, "\"ClearQuest Integration\"", $branchindex);
        if ( $triggerindex != -1 ) {
            # Get the next trigger type block's starting point
            $branchindex = index($vobresult, "\"ClearQuest Integration\"", $triggerindex+1);
            # Get one trigger type block
            if ( $branchindex != -1 ) {
                $onetrigger = substr $vobresult, $triggerindex, $branchindex-$triggerindex;
            } else {
                # Last trigger type block
                $onetrigger = substr $vobresult, $triggerindex;
                $checktriggers = 0;
            }
            # Look for any branch type restrictions on the current trigger type
            $brtypecount = () = $onetrigger =~ /restriction: /g;
            if ( $brtypecount == 0 ) {
                # No restrictions means all branch types have this trigger type
                # So all branch types will be migrated
                $checktriggers = 0;
                %branchtypes = ();
                $branchtypes{"All Branch Types"} = 1;
            } else {
                $tempbrpos = 0;
                for ( my $k = 0; $k < $brtypecount; $k++ ) {
                    # For each restriction listed for a branch type...
                    $brtype = "restriction: branch type ";
                    $tempbrpos = index($onetrigger, $brtype, $tempbrpos+1);
                    # Get the end of the line for the current branch type restriction
                    $endofbrtype = index($onetrigger, "\n", $tempbrpos);
                    # Get the name of the branch type
                    $brtype = substr $onetrigger, $tempbrpos+(length $brtype), ($endofbrtype - ($tempbrpos+(length $brtype)));
                    # Add the branch type to the list
                    $branchtypes{$brtype} = 1;
                }
            }
        } else {
            die "\nNo ClearQuest trigger types found.";
	}
    }
}

# Save the CQ information to file
my $tempinput = "";

my $usescript = "";
print "\nDo you wish to enter CMI configuration for each brtype now? [y] \nEnter 'n' to enter the information manually later.";
$usescript = <STDIN>;
chomp $usescript;

open my $tempfile, ">>", "$migration_file_name" or die $!;
foreach my $currentdb ( @dblist ) {
    foreach my $currentbrtype ( keys %branchtypes ) {
	# For each DB write the details we know already
	print $tempfile "BrType:$currentbrtype\n";
	print $tempfile "DB:$currentdb\n";
	print $tempfile "DBSet:$dbset\n";
	print $tempfile "Server URL:$serverurl\n";
	# Type is assumed to be CQ
	print $tempfile "Type:cmcq\n";
	# Assume v1_0 since it is the only existing version
	print $tempfile "Version:v1_0\n";

	if ( $usescript ne "n" ) {
	    print "\nPlease enter the remaining information for:";
	    print "\nBrType:$currentbrtype";
	    print "\nDB:$currentdb";
	    print "\nDBSet:$dbset";
	    print "\nServer URL:$serverurl";
	    # Prompt the user to provide each missing piece of information or manually fill in the values

	    # Get the description
	    print "\nEnter the description (e.g. \"CQ Web Server 1\"):";
	    $tempinput = <STDIN>;
	    chomp $tempinput;
	    exit 0 if ( $tempinput eq "" );
	    print $tempfile "Description:$tempinput\n";
	    $tempinput = "";

	    # Get the queryuri
	    print "\nEnter the query URI (e.g. \"rcm.name=Public%20Queries/All%20Defects\") \nor return to enter no query URI:";
	    $tempinput = <STDIN>;
	    chomp $tempinput;
	    # Query URI can be empty
	    print $tempfile "Query:$tempinput\n";
	    $tempinput = "";

	    # Get the provider name
	    print "\nEnter the Provider Name (e.g. \"CQPROV1\"):";
	    $tempinput = <STDIN>;
	    chomp $tempinput;
	    exit 0 if ( $tempinput eq "" );
	    print $tempfile "Name:$tempinput\n";
	    $tempinput = "";
	} else {
            # Manually filling in values later
            print $tempfile "Description:\n";
       	    print $tempfile "Query:\n";
	    print $tempfile "Name:\n";
	}
	print $tempfile "\n";
    }
}

close $tempfile;

# User has chosen to not continue with setup for now, exit
if ( $usescript eq "n" ) {
    print "\nPlease fill in the required details in $migration_file_name and run the script again to resume.";
    exit 0;
}

} # End of sub setup


# This does the actual migration of the perl integration to CMI
sub migrate {
open (migration_file, "<", "$migration_file_name") or die $!;

my @brtypes;
my @DBs;
my @DBSets;
my @provversions;
my @descs;
my @types;
my @urls;
my @provnames;
my @queries;

my $temp = 0;
my $tempstr = "";

# Grab the information from the file for each provider to setup
print "\nReading from $file_name...";
while ( my $migrationline = <migration_file> ) {
    # Remove leading and trailing whitespace from $line
    $migrationline =~ s/^\s+|\s+$//g;
    # Find the first ':' in the line
    $temp = index($migrationline, ":");
    if ( $migrationline =~ /BrType:/ ) {
        $tempstr = substr $migrationline, $temp+1;
	# Trim whitespace
	$tempstr =~ s/^\s+|\s+$//g;
        if ( $tempstr ne "" ) {
    	    push(@brtypes, $tempstr);
        }
    } elsif ( $migrationline =~ /DB:/ ) {
        $tempstr = substr $migrationline, $temp+1;
	# Trim whitespace
	$tempstr =~ s/^\s+|\s+$//g;
        if ( $tempstr ne "" ) {
    	    push(@DBs, $tempstr);
        }
    } elsif ( $migrationline =~ /DBSet:/ ) {
    	$tempstr = substr $migrationline, $temp+1;
	# Trim whitespace
	$tempstr =~ s/^\s+|\s+$//g;
        if ( $tempstr ne "" ) {
    	    push(@DBSets, $tempstr);
        }
    } elsif ( $migrationline =~ /Version:/ ) {
    	$tempstr = substr $migrationline, $temp+1;
	# Trim whitespace
	$tempstr =~ s/^\s+|\s+$//g;
        if ( $tempstr ne "" ) {
    	    push(@provversions, $tempstr);
        }
    } elsif ( $migrationline =~ /Description:/ ) {
        $tempstr = substr $migrationline, $temp+1;
	# Trim whitespace
	$tempstr =~ s/^\s+|\s+$//g;
	# If the description is not enclosed in double quotes we need to add them
	if ( index($tempstr, "\"") != 0 ) {
	    $tempstr = "\"" . $tempstr . "\"";
	}
        if ( $tempstr ne "" ) {
    	    push(@descs, $tempstr);
        }
    } elsif ( $migrationline =~ /Type:/ ) {
    	$tempstr = substr $migrationline, $temp+1;
	# Trim whitespace
	$tempstr =~ s/^\s+|\s+$//g;
        if ( $tempstr ne "" ) {
    	    push(@types, $tempstr);
        }
    } elsif ( $migrationline =~ /URL:/ ) {
    	$tempstr = substr $migrationline, $temp+1;
	# Trim whitespace
	$tempstr =~ s/^\s+|\s+$//g;
        if ( $tempstr ne "" ) {
    	    push(@urls, $tempstr);
        }
    } elsif ( $migrationline =~ /Name:/ ) {
    	$tempstr = substr $migrationline, $temp+1;
	# Trim whitespace
	$tempstr =~ s/^\s+|\s+$//g;
        if ( $tempstr ne "" ) {
    	    push(@provnames, $tempstr);
        }
    } elsif ( $migrationline =~ /Query:/ ) {
    	$tempstr = substr $migrationline, $temp+1;
	# Trim whitespace
	$tempstr =~ s/^\s+|\s+$//g;
        # Queries aren't required so it doesn't matter if it's empty
	if ( index($tempstr, "\"") != 0 ) {
	    $tempstr = "\"" . $tempstr . "\"";
	}
    	push(@queries, $tempstr);
    }
    # Any other data are ignored
}
close migration_file;

# Is all of the required information there?
# If yes, continue
# If no, quit (can resume from first function to fill in)
if ( !grep { $_ != (scalar @brtypes) } (scalar @DBs, scalar @DBSets, scalar @provversions, scalar @descs, scalar @types, scalar @urls, scalar @provnames) ) {
    print "\n$file_name information verified.";
} else {
    die "\nInformation missing from $file_name.";
} 

print "\nMaking attypes in VOB...";

# Setup attypes
# Check if the attypes already exist
# If they exist don't try making them again
my $attcheck = `cleartool lstype -kind attype -invob $vobinput 2>&1`;
if ( $attcheck =~ /cleartool: / && ${^CHILD_ERROR_NATIVE} != 0 ) {
    die "\nError checking attypes in $vobinput.";
}
if ( !($attcheck =~ /"CC_CMI_CONTEXT"/) ) {
    $vobresult = `cleartool mkattype -shared -nc -vtype string CC_CMI_CONTEXT\@$vobinput 2>&1`;
    if ( $vobresult =~ /cleartool: / && ${^CHILD_ERROR_NATIVE} != 0 ) {
    	die "\nError making attype in $vobinput.";
    }
}
if ( !($attcheck =~ /"CC_CMI_PROVIDERS"/) ) {
    $vobresult = `cleartool mkattype -shared -nc -vtype string CC_CMI_PROVIDERS\@$vobinput 2>&1`;
    if ( $vobresult =~ /cleartool: / && ${^CHILD_ERROR_NATIVE} != 0 ) {
	die "\nError making attype in $vobinput.";
    }
}
if ( !($attcheck =~ /"CC_CMI_TASK"/) ) {
    $vobresult = `cleartool mkattype -shared -nc -vtype string CC_CMI_TASK\@$vobinput 2>&1`;
    if ( $vobresult =~ /cleartool: / && ${^CHILD_ERROR_NATIVE} != 0 ) {
	die "\nError making attype in $vobinput.";
    }
}

print "\nRunning mkcmprovider for VOB and branch types...";
my $tempbrtype = "";
my $tempbrtypepos = 0;
my $tempbrtypepos2 = 0;
my $brtypevar = "branch type \"";
my $allbrtypes = "All Branch Types";
my $brtypecount = 0;
my $tempresult = "";
my $tempquerystr = "";
my $providererror = 0;
my %providerhash;
my @providerarray = ();
my $skipvob = 0;
foreach my $j ( 0 .. $#DBs ) {
    # Run mkcmprovider on VOB
	
    $skipvob = 0;
    # If we've already run the same mkcmprovider command on this VOB we don't
    # want to run it again because it will just throw an error
    # Find out if the current provider was already done before
    for ( my $m = $j; $m < (scalar @DBs); $m++ ) {
	# See if the Provider Name exists in the hash
	if ( exists($providerhash{$provnames[$m]}) ) {
	    # Provider Name mkcmprovider has already been run
	    # If the rest of the details match let's skip it
	    # Otherwise assume it's an error and go ahead so the
	    # user is aware
	    @providerarray = exists( $providerhash{$provnames[$m]} ) ? @{$providerhash{$provnames[$m]}} : ();
	    # Check if they contain the same members
	    if ( ($provversions[$m] eq $providerarray[0]) && ($descs[$m] eq $providerarray[1]) && ($types[$m] eq $providerarray[2]) && ($urls[$m] eq $providerarray[3]) ) {
		# They're the same so we should skip the next mkcmprovider -vob command
		$skipvob = 1;
		last;
	    } else {
		$skipvob = 0;
	    }
	}
    }
    if ( !$skipvob ) {
    	$vobresult = `cleartool mkcmprovider -vob vob:$vobinput -version $provversions[$j] -description $descs[$j] -type $types[$j] -connection baseurl:$urls[$j] $provnames[$j] 2>&1`;
	if ( $vobresult =~ /cleartool: / && ${^CHILD_ERROR_NATIVE} != 0 ) {
	    print "\nError running mkcmprovider -vob.\n$vobresult";
	    $providererror = 1;
   	}
	# Keep track of the Provider we added so we don't try to add it again
  	@providerarray = ();
  	push (@providerarray, $provversions[$j]);
  	push (@providerarray, $descs[$j]);
  	push (@providerarray, $types[$j]);
  	push (@providerarray, $urls[$j]);
  	$providerhash{$provnames[$j]} = \@providerarray;
    }
    # We need to either have a queryuri or not specify -queryuri so
    # if the queryuri is empty we want to just not include it at all (since it is optional)
    $tempquerystr = $queries[$j];
    # Use 2 as the cutoff since empty will be two double quotes here
    if ( length($tempquerystr) > 2 ) {
	$tempquerystr = ",queryuri:$tempquerystr";
    } else {
	$tempquerystr = "";
    }
    # Check if @brtypes' current element is "All Branch Types" -> means we need to find all the branch types
    # in the VOB and enable each one for this DB
    if ( $brtypes[$j] eq $allbrtypes ) {
	$vobresult = `cleartool lstype -kind brtype -invob $vobinput 2>&1`;
	if ( $vobresult =~ /cleartool: / && ${^CHILD_ERROR_NATIVE} != 0 ) {
	    die "\nError getting branch types in VOB $vobinput.";
   	}
	# If there's at least one branch type in the VOB
	$brtypecount = () = $vobresult =~ /$brtypevar/g;
	if ( $brtypecount > 0 ) {
	    # Run mkcmprovider for each brtype
	    for ( my $l = 0; $l < $brtypecount; $l++ ) {
		$tempbrtypepos = (index($vobresult, $brtypevar, $tempbrtypepos) + length $brtypevar);
		$tempbrtypepos2 = index($vobresult, "\"", $tempbrtypepos);
		$tempbrtype = substr $vobresult, $tempbrtypepos, ($tempbrtypepos2 - $tempbrtypepos);
		$tempresult = `cleartool mkcmprovider -brtype $tempbrtype\@$vobinput -context userdb:$DBs[$j],dbset:$DBSets[$j]$tempquerystr $provnames[$j] 2>&1`;
		if ( $tempresult =~ /cleartool: / && ${^CHILD_ERROR_NATIVE} != 0 ) {
		    print "\nError making CM provider for branch type: $tempbrtype.\n$tempresult";
		    $providererror = 1;
	    	}
	    }
	} else {
	    die "\nDid not find any branch types in VOB.";
	}
    } else {
	# @brtype's current element is a specific branch type
	$vobresult = `cleartool mkcmprovider -brtype $brtypes[$j]\@$vobinput -context userdb:$DBs[$j],dbset:$DBSets[$j]$tempquerystr $provnames[$j] 2>&1`;
	if ( $vobresult =~ /cleartool: / && ${^CHILD_ERROR_NATIVE} != 0 ) {
	    print "\nError making CM provider for branch type: $brtypes[$j].\n$vobresult";
	    $providererror = 1;
	}
    }
}

if ( $providererror != 0 ) {
    print "\nThere were errors creating providers.";
    print "\nPlease check to see if the desired providers were created correctly or run this script again.";
}

# At this point the providers should be setup completely on the VOB and branch types
# Now we need to grab the hlinks from the versions and store them in a format 
# that the repair tool (clmutil) can use to make them into CMI associations

open my $verfile, ">>", "$migration_versions_file_name" or die $!;

# Find all versions with ClearQuest hyperlinks in the VOB
my @versions;
my $versionstr = "";
my $versionstr2 = "";
my $curtask = "";
my $curdb = "";
my $curview = `cleartool pwv -s 2>&1`;
my $temppos = 0;
my @hyperlinks;
my $versionerror = 0;

my %dbhash;
# Build a hash of DBs and their position in the DBs array
# Since the perl integration only supports on DBSet there should NOT
# be any DBs with the same names.  Multiple occurrences of a DB name
# will be overwritten with whatever is hit last.
for ( my $m = 0; $m < (scalar @DBs); $m++ ) {
    $dbhash{$DBs[$m]} = $m;
}

if ( ($curview =~ /cleartool: / && ${^CHILD_ERROR_NATIVE} != 0) || $curview =~ /\*\* / ) {
    die "\nError retrieving current view.";
}
@versions = `cleartool find $vobinput -version \"hltype\(CrmRequest\)\" -print 2>&1`;

if ( (scalar @versions) == 0 || ($versions[0] =~ /cleartool: / && ${^CHILD_ERROR_NATIVE} != 0) ) {
    die "\nError retrieving versions with ClearQuest hyperlinks.";
}
# For each version found build up the version string required by the
# repair tool (clmutil)
# e.g.
# {"ACTION_KIND":"Associate","CCRES_DISP_NAME":"\cmi_vob\test.c@@\main\CHECKEDOUT",
# "CCRES_MTYPE":"8","CCRES_USEL":"oid:1a26ae970aa54e439ead588c10020033@vobuuid:3474538a35514cecbd0317bd4f63d235",
# "CC_VIEW_TAG":"cmi_view","CM_RESOURCE_ID":"CQWAN00036486@CQPROV"}
# 
# MTYPE = 8 for versions
foreach my $curver ( @versions ) {
    $versionstr = "\{\"ACTION_KIND\":\"Associate\",\"CCRES_DISP_NAME\":\"";
    chomp $curver;
    $versionstr .= $curver;
    $versionstr .= "\",\"CCRES_MTYPE\":\"8\",\"CCRES_USEL\":\"";
    # Leave the oid@vobuuid blank.  Let the repair tool (clmutil) figure it out.
    $versionstr .= "\",";
    $versionstr .= "\"CC_VIEW_TAG\":\"";
    chomp $curview;
    $versionstr .= $curview . "\",";
    $versionstr .= "\"CM_RESOURCE_ID\":\"";
    # Get the CrmRequest hyperlinks on the current version
    @hyperlinks = `cleartool desc -ihlink CrmRequest $curver 2>&1`;
    if ( $hyperlinks[0] =~ /cleartool: / && ${^CHILD_ERROR_NATIVE} != 0 ) {
	print "\nError retrieving hyperlinks for $curver.";
	$versionerror = 1;
	# Couldn't get the hyperlinks for this version
	# Go to next version
	next;
    }
    foreach my $task ( @hyperlinks ) {
	# Trim whitespace
	$task =~ s/^\s+|\s+$//g;
	if ( index($task, "CrmRequest") == 0 ) {
	    # If the hyperlink is a CrmRequest get the task ID from it
	    # e.g. SAMPL00000040
	    $temppos = index($task,"\"",(index($task,"-> "))) + 1;
	    $curtask = substr $task, $temppos, (index($task,"\"",$temppos) - $temppos);
	    $versionstr2 = $versionstr . $curtask;
	    # Extract the DB name from the task ID
	    # CQ ID #s are 8 digits
	    $curdb = substr $curtask, 0, (length($curtask) - 8);
	    # Add the Provider Name for the given DB
	    $versionstr2 .= "\@" . $provnames[$dbhash{$curdb}] . "\"\}";
	    # Add the association to the file
	    print $verfile "$versionstr2\n";
	}
    }
}

if ( $versionerror != 0 ) {
    print "\nThere were errors retrieving hyperlinks for some versions.";
    print "\nPlease check to see if the correct versions were recorded or run this script again.";
}

print "\nMigration has completed successfully.\nPlease run the CMI Repair Tool (clmutil) to link versions to ClearQuest records using the migration version information file.\n";

} # End of sub migrate



# Test for the existence of HKEY_LOCAL_MACHINE\Software\Wow6432Node
sub winbitcheck {
    if ( $^O =~ /Win32/ ) {
	my $registercheck = "HKEY_LOCAL_MACHINE\\Software\\Wow6432Node";
	my $bitReturn;

	if ( $Registry->{$registercheck} ) {
	    $bitReturn = "64";
	}
	else {
	    $bitReturn = "32"
    	}
    	return $bitReturn;
    }
} # End of sub winbitcheck

